#include "gbsplinesurface.h"

namespace GraphicsObjects {

    GBSplineSurface::GBSplineSurface(std::string name, Point * pointsMatrix[4][4], Color * color) : GCurve(name, color)
    {
        for (int line = 0; line < 4; line ++) {
            for (int column = 0; column < 4; column ++) {
                GBSplineSurface::pointsMatrix[line][column] = pointsMatrix[line][column]->copy();
                GBSplineSurface::points.push_back(GBSplineSurface::pointsMatrix[line][column]);
            }
        }
    }

    GBSplineSurface * GBSplineSurface::copy() {
        return new GBSplineSurface(name, pointsMatrix, color);
    }

    std::string GBSplineSurface::type(){
        return "B-Spline Surface";
    }

    std::list<Point*>  GBSplineSurface::generatePoints(float desiredAccuracy){
        std::list<Point*> generatedPoints;

        //prepare matrixes
        float geometryMatrix[4][4][4];
        for (int line = 0; line < 4; line ++) {
            for (int column = 0; column < 4; column ++) {
                pointsMatrix[line][column]->toVector(geometryMatrix[line][column]);
            }
        }
        float geometryMatrixOrderedByCoordinates[3][4][4];
        for (int coordinate = 0; coordinate < 3; coordinate ++) {
            for (int line = 0; line < 4; line ++) {
                for (int column = 0; column < 4; column ++) {
                    geometryMatrixOrderedByCoordinates[coordinate][line][column] = geometryMatrix[line][column][coordinate];
                }
            }
        }
        float methodMatrix[4][4] = {{-1./6,  1./2, -1./2, 1./6},
                                    { 1./2, -1.,    1./2, 0.  },
                                    {-1./2,  0.,    1./2, 0.  },
                                    { 1./6,  2./3,  1./6, 0.  }};

        //determine step based on desired accuracy
        float bezierMatrix[4][4] = {{-1.,  3., -3., 1.},
                                    { 3., -6.,  3., 0.},
                                    {-3.,  3.,  0., 0.},
                                    { 1.,  0.,  0., 0.}};
        float accuracy = INFINITY;
        float step = MAXIMUM_STEP * 2;
        while (accuracy > desiredAccuracy && step > MINIMUM_STEP) {
            step = step/2;
            Point * tmp = new Point();
            float tVector[4] = {step*step*step, step*step, step, 1};
            float tempVector[4];
            Transformations::multiply(tVector, bezierMatrix, tempVector);
            float coordinates[4];
            Transformations::multiply(tempVector, geometryMatrix[0], coordinates);
            tmp->fromVector(coordinates);
            accuracy = pointsMatrix[0][0]->distance(tmp);
        }
        if (step < MINIMUM_STEP)
            step = MINIMUM_STEP;

        int n = 1 / step;

        //calculate composed matrix
        float composedMatrix[3][4][4];
        float methodMatrixTranspose[4][4];
        Transformations::transpose(methodMatrix, methodMatrixTranspose);
        for (int i = 0; i < 3; i++) {
            Transformations::multiply(methodMatrix, geometryMatrixOrderedByCoordinates[i], composedMatrix[i]);
            Transformations::multiply(composedMatrix[i], methodMatrixTranspose, composedMatrix[i]);
        }

        Point * matrixOfPoints[n][n];
        float s = 0;

        for (int j = 0; j < n; j++) {

            //calculate initial values of curve
            float sVector[4] = {s*s*s, s*s, s, 1};
            float tmpVector[4], curveVector[4];
            float f[3], df[3], ddf[3], dddf[3];
            for (int i = 0; i < 3; i++) {
                //Transformations::multiply(sVector, composedMatrix[i], curveVector);
                Transformations::multiply(sVector, methodMatrix, curveVector);
                Transformations::multiply(curveVector, geometryMatrixOrderedByCoordinates[i], tmpVector);
                Transformations::multiply(tmpVector, methodMatrixTranspose, curveVector);
                f[i] = curveVector[3];
                df[i] = curveVector[0]*step*step*step + curveVector[1]*step*step + curveVector[2]*step;
                ddf[i] = 6*curveVector[0]*step*step*step + 2*curveVector[1]*step*step;
                dddf[i] = 6*curveVector[0]*step*step*step;
            }

            for (int i = 0; i < n; i++){
                Point * tmp = new Point();

                tmp->fromVector(f);
                for (int k = 0; k < 3; k++) {
                    f[k] += df[k];
                    df[k] += ddf[k];
                    ddf[k] += dddf[k];
                }

                matrixOfPoints[j][i] = tmp;
            }

            s += step;

        }

        for (int line = 0; line < n; line ++) {
            for (int column = 0; column < n-1; column ++) {
                generatedPoints.push_back(matrixOfPoints[line][column]);
                generatedPoints.push_back(matrixOfPoints[line][column+1]);
            }
        }
        for (int line = 0; line < n-1; line ++) {
            for (int column = 0; column < n; column ++) {
                generatedPoints.push_back(matrixOfPoints[line][column]);
                generatedPoints.push_back(matrixOfPoints[line+1][column]);
            }
        }

        return generatedPoints;
    }

    bool GBSplineSurface::clip(float accuracy) {
        std::list<Point*> generatedPoints = generatePoints(accuracy);
        std::list<Point*>::iterator it = generatedPoints.begin();

        points.clear();
        Point * p1;
        Point * p2;

        while(it != generatedPoints.end()){
            p1 = *it++;
            p2 = *it++;

            std::list<Point*> clippedLine = clipLine(p1->copy(), p2->copy());
            points.splice(points.end(), clippedLine);

        }

        return points.size() > 0;
    }

} // namespace GraphicsObjects
