#include "transformations.h"

namespace Transformations {

    void getTranslationMatrix(Point * translationVector, float resultMatrix[4][4]){
        float dx = translationVector->x;
        float dy = translationVector->y;
        float dz = translationVector->z;
        float matrix[4][4] = {{1,0,0,0},{0,1,0,0},{0,0,1,0},{dx,dy,dz,1}};
        copy(matrix, resultMatrix);
    }

    void getScaleMatrix(Point * anchor, Point * scaleVector, float resultMatrix[4][4]) {
        float dx = anchor->x;
        float dy = anchor->y;
        float dz = anchor->z;
        float sx = scaleVector->x;
        float sy = scaleVector->y;
        float sz = scaleVector->z;
        float matrix[4][4] = {{sx,0,0,0},{0,sy,0,0},{0,0,sz,0},{-dx*sx+dx, -dy*sy+dy, -dz*sz+dz, 1}};
        copy(matrix, resultMatrix);
    }

    void getRotationMatrix(Axis axis, float angle, float resultMatrix[4][4]) {
        float rad = angle * M_PI/180;
        float s = sin(rad);
        float c = cos(rad);
        if (axis == X_AXIS) {
            float matrix[4][4] = {{1, 0, 0, 0},
                                  {0, c, s, 0},
                                  {0,-s, c, 0},
                                  {0, 0, 0, 1}};
            copy(matrix, resultMatrix);
        } else if (axis == Y_AXIS) {
            float matrix[4][4] = {{c, 0,-s, 0},
                                  {0, 1, 0, 0},
                                  {s, 0, c, 0},
                                  {0, 0, 0, 1}};
            copy(matrix, resultMatrix);
        } else if (axis == Z_AXIS) {
            float matrix[4][4] = {{c, s, 0, 0},
                                 {-s, c, 0, 0},
                                  {0, 0, 1, 0},
                                  {0, 0, 0, 1}};
            copy(matrix, resultMatrix);
        }
    }

    void getTransformationMatrix(Point * translationVector, Point * scaleVector, GPoint * vpn, GPoint * vpu, float resultMatrix[4][4]) {
        Point * angles = findAlignmentAngles(vpn, vpu);
        float matrixA[4][4], matrixB[4][4], matrixC[4][4];
        getTranslationMatrix(translationVector, matrixA);
        getRotationMatrix(X_AXIS, angles->x, matrixB);
        multiply(matrixA, matrixB, matrixC);
        getRotationMatrix(Y_AXIS, angles->y, matrixB);
        multiply(matrixC, matrixB, matrixA);
        getRotationMatrix(Z_AXIS, angles->z, matrixB);
        multiply(matrixA, matrixB, matrixC);
        getScaleMatrix(new Point(0, 0, 0), scaleVector, matrixB);
        multiply(matrixC, matrixB, resultMatrix);
    }

    int translate(GObject * object, float dx, float dy, float dz){
        float matrix[4][4];
        getTranslationMatrix(new Point(dx, dy, dz), matrix);
        return transform(object, matrix);
    }

    int translate(GObject * object, Point * axisVector, float distance) {
        normalize(axisVector);

        return translate(object, axisVector->x * distance, axisVector->y * distance, axisVector->z * distance);
    }

    int rotate(GObject * object, Axis axis, float angle){
        float matrix[4][4];
        getRotationMatrix(axis, angle, matrix);
        return transform(object, matrix);
    }

    int rotate(GObject * object, Point * anchor, Point * axisVector, float angle) {

        if (axisVector->x == 0 && axisVector->y == 0 && axisVector->z == 0) {
            return ERROR_NULL_VECTOR;
        }

        normalize(axisVector);

        float ux = axisVector->x;
        float uy = axisVector->y;
        float uz = axisVector->z;
        float ux2 = ux*ux;
        float uy2 = uy*uy;
        float uz2 = uz*uz;
        float rad = angle * M_PI/180;
        float c = cos(rad);
        float s = sin(rad);
        float dx = anchor->x;
        float dy = anchor->y;
        float dz = anchor->z;

        float matrix[4][4] =
            {{ux2+(1-ux2)*c,    ux*uy*(1-c)-uz*s, ux*uz*(1-c)+uy*s, 0},
             {ux*uy*(1-c)+uz*s, uy2+(1-uy2)*c,    uy*uz*(1-c)-ux*s, 0},
             {ux*uz*(1-c)-uy*s, uy*uz*(1-c)+ux*s, uz2+(1-uz2)*c,    0},
             {0,                0,                0,                1}};
        matrix[3][0] = -dx*matrix[0][0] -dy*matrix[1][0] -dz*matrix[2][0] +dx;
        matrix[3][1] = -dx*matrix[0][1] -dy*matrix[1][1] -dz*matrix[2][1] +dy;
        matrix[3][2] = -dx*matrix[0][2] -dy*matrix[1][2] -dz*matrix[2][2] +dz;

        return transform(object, matrix);
    }

    int scale(GObject * object, float sx, float sy, float sz){
        return scale(object, new Point(0, 0, 0), new Point(sx, sy, sz));
    }

    int scale(GObject * object, Point * anchor, Point * scaleVector){
        float matrix[4][4];
        getScaleMatrix(anchor, scaleVector, matrix);
        return transform(object, matrix);
    }

    int transform(GObject * object, float matrix[4][4]) {
        std::list<Point*> listpoints = object->getPoints();
        if(listpoints.size()==0)
            return ERROR_AMOUNT_OF_POINTS;

        std::list<Point*>::iterator it;
        float vet[4] = {0,0,0,1};
        float resultVector[4];
        for (it = listpoints.begin(); it != listpoints.end(); it++){
            Point *p = (*it);
            vet[0] = p->x;
            vet[1] = p->y;
            vet[2] = p->z;
            multiply(vet, matrix, resultVector);
            p->x = resultVector[0];
            p->y = resultVector[1];
            p->z = resultVector[2];
        }

        return OK;
    }

    void transform(Point * point, float matrix[4][4]) {
        float vector[4];
        float resultVector[4];
        point->toVector(vector);
        multiply(vector, matrix, resultVector);
        point->fromVector(resultVector);
    }

    std::list<GObject*> transform(std::list<GObject*> originalObjects, Point * translationVector, Point * scaleVector, float distanceFromProjectionCenter, GPoint * vpn, GPoint * vpu) {
        float matrix[4][4];
        getTransformationMatrix(translationVector, scaleVector, vpn, vpu, matrix);
        float normalizedDistanceFromProjectionCenter = distanceFromProjectionCenter * scaleVector->z;
        std::list<GObject*> transformedObjects;
        std::list<GObject*>::iterator objectsIt;
        for (objectsIt = originalObjects.begin(); objectsIt != originalObjects.end(); objectsIt++) {
            GObject * object = (*objectsIt)->copy();
            std::list<Point*> pointsList = object->getPoints();
            std::list<Point*>::iterator pointsIt;
            for (pointsIt = pointsList.begin(); pointsIt != pointsList.end(); pointsIt++){
                transform(*pointsIt, matrix);
                (*pointsIt)->x *= normalizedDistanceFromProjectionCenter / ((*pointsIt)->z + normalizedDistanceFromProjectionCenter);
                (*pointsIt)->y *= normalizedDistanceFromProjectionCenter / ((*pointsIt)->z + normalizedDistanceFromProjectionCenter);

            }
            transformedObjects.push_back(object);
        }
        return transformedObjects;
    }

    bool normalize(Point * vector) {
        float vectorNorm = sqrt(vector->x*vector->x + vector->y*vector->y + vector->z*vector->z);
        if (vectorNorm == 0) {
            return false;
        }
        vector->x /= vectorNorm;
        vector->y /= vectorNorm;
        vector->z /= vectorNorm;
        return true;
    }

    Point * crossProduct (Point * u, Point * v) {
        return new Point(u->y*v->z-u->z*v->y, u->z*v->x-u->x*v->z, u->x*v->y-u->y*v->x);
    }

    Point * findAlignmentAngles(GPoint * vpnOriginal, GPoint * vpuOriginal) {
        GPoint * vpn = vpnOriginal->copy();
        GPoint * vpu = vpuOriginal->copy();

        //puts vpn on xz plane
        Point * yz = vpn->point()->copy();
        yz->x = 0;
        float angleX = 0;
        if (normalize(yz)) {
            angleX = atan2(yz->y, yz->z) * 180/M_PI;
            rotate(vpn, X_AXIS, angleX);
            rotate(vpu, X_AXIS, angleX);
        }

        //aligns vpn with z axis
        float angleY = - atan2(vpn->point()->x, vpn->point()->z) * 180/M_PI;
        rotate(vpn, Y_AXIS, angleY);
        rotate(vpu, Y_AXIS, angleY);

        //aligns vpu with y axis
        float angleZ = atan2(vpu->point()->x, vpu->point()->y) * 180/M_PI;
        rotate(vpn, Z_AXIS, angleZ);
        rotate(vpu, Z_AXIS, angleZ);

        return new Point(angleX, angleY, angleZ);
    }

    void multiply(float matrixA[4][4], float matrixB[4][4], float resultMatrix[4][4]) {
        for (int line = 0; line < 4; line++) {
            for (int column = 0; column < 4; column++) {
                resultMatrix[line][column] = 0;
                for (int element = 0; element < 4; element++) {
                    resultMatrix[line][column] += matrixA[line][element] * matrixB[element][column];
                }
            }
        }
    }

    void multiply(float lineVector[4], float matrix[4][4], float resultVector[4]) {
        for (int column = 0; column < 4; column++) {
            resultVector[column] = 0;
            for (int element = 0; element < 4; element++) {
                resultVector[column] += lineVector[element] * matrix[element][column];
            }
        }
    }

    void multiply(float matrix[4][4], float columnVector[4], float resultVector[4]) {
        for (int line = 0; line < 4; line++) {
            resultVector[line] = 0;
            for (int element = 0; element < 4; element++) {
                resultVector[line] += matrix[line][element] * columnVector[element];
            }
        }
    }

    void transpose(float matrix[4][4], float resultMatrix[4][4]) {
        for (int line = 0; line < 4; line++) {
            for (int column = 0; column < 4; column++) {
                resultMatrix[column][line] = matrix[line][column];
            }
        }
    }

    void copy(float originalMatrix[4][4], float copyMatryx[4][4]) {
        for (int line = 0; line < 4; line++) {
            for (int column = 0; column < 4; column++) {
                copyMatryx[line][column] = originalMatrix[line][column];
            }
        }
    }

}
